/*********************************************************************
 * Authors: B. Alex Bridges and Miriam Becker                        *
 * File: prodcons.cc/threadtest.cc                                   *
 * Class: CPSC-451, Summer 1999                                      *
 * Project: Project 4 - Producer-Consumer Problem Solved via         *
 *                      Monitors (Implemented Using Semaphores)      *
 * Description: This program has a producer which places characters  *
 *              from a file into a buffer, and a consumer which pulls*
 *              characters ut of the buffer abd prints them to the   *
 *              screen.                                              *
 * Contents: Methods for creating and controlling the producers and  *
 *           consumers as well as the actual code for the producers  *
 *           and consumers.                                          *
 *********************************************************************/
 
/* INCLUDES */
#include "copyright.h"
#include "system.h"
#include "synch.h"
#include "stdio.h"
#include "stdlib.h"

/* CONSTANTS */
// NONE

/* GLOBAL VARIABLES */
FILE* FileInput = NULL;    // POINTER TO INPUT FILE
char* charABuffer;         // BUFFER OF READ CHARACTERS
int intBufferSize = 0;     // SIZE OF BUFFER
int intCIndex = 0;         // CONSUMER/WRITE INDEX INTO BUFFER
int intPIndex = 0;         // PRODUCER/READ INDEX INTO BUFFER
int intCharCount = 0;      // NUMBER OF NEW CHARACTERS IN THE BUFFER
int intEOF = 0;            // END-OF-FILE STATUS WHERE 1 = REACHED AND 0 = NOT REACHED
int next_count = 0;        // NUMBER OF PROCEDURES THAT ARE IN MONITOR, BUT WAITING
// => CONDITIONS
Condition* full;           // CONDITION FOR FULLNESS OF BUFFER
Condition* empty;          // CONDITION FOR EMPTINESS OF BUFFER
// => SEMAPHORES
Semaphore* file;           // SEMAPHORE FOR FILE CONTROL
Semaphore* mutex;          // SEMAPHORE FOR MUTEX CONTROL
Semaphore* next;           // SEMAPHORE FOR NEXT CONTROL


/* METHODS */
/*************************************************************************
 * Method: Producer                                                      *
 * Purpose: Reads a character from the item referenced by the file       *
 *          pointer and places it in a rotating buffer.                  *
 * Input: --PARAMETERS--                                                 *
 *        => which = ID value for the running thread.                    *
 * Output: --RETURNS--                                                   *
 *        => none                                                        *
 *************************************************************************/
void Producer(int which)
{
  /* LOCAL VARIABLES */
  char charNextP;  // CHARACTER PRODUCED/READ
  
  do
  {
    /* CHECK END-OF-FILE STATUS */
    if(!intEOF)
    {
      /* GET CHARACTER FROM FILE */
      file->P();
      charNextP = fgetc(FileInput);
      file->V();

      /* CHECK CHARACTER FROM FILE */
      // => BAD CHARACTER
      if(charNextP == EOF)
      {
        /* SIGNAL END OF FILE FLAG AND CLOSE FILE */
        intEOF = 1;
        fclose(FileInput);
      } // if
      // => GOOD CHARACTER
      else
      {
        /* MONITOR ENTRY */
        DEBUG('c', " Producer '%dp' waiting on mutex.\n\n", which);
        mutex->P();

        /* CHECK FULLNESS OF BUFFER */
        DEBUG('c', " Producer '%dp' found Character Count vs. Buffer Size was '%d' vs. '%d'.\n\n", which, intCharCount,intBufferSize);
        if(intCharCount == intBufferSize)
        {
          DEBUG('c', " Producer '%dp' conditionally waiting on full.\n\n", which);
          full->Wait();
        } // if

        /* PUT CHARACTER IN BUFFER */
        charABuffer[intPIndex] = charNextP;
        printf("Producer '%dp' producing '%c'\n", which, charNextP);
        DEBUG('c', " The index written to was '%d'.\n\n", intPIndex);

        /* ADVANCE INDEX */
        intPIndex = (intPIndex + 1) % intBufferSize;

        /* INC. NEW CHARACTER COUNT */
        intCharCount++;

        /* SIGNAL CONSUMER */
        DEBUG('c', " Producer '%dp' conditionally signaling consumer via empty.\n\n", which);
        empty->Signal();

        /* MONITOR EXIT */
        if (next_count > 0)
        {
          DEBUG('c', " Producer '%dp' signaling next.\n\n", which);
          next->V();
        } // if
        else
        {
          DEBUG('c', " Producer '%dp' signaling mutex.\n\n", which);
          mutex->V();
        } // else
      } // else
    } // while
    else
    {
      DEBUG('c', " Producer '%dp' found EOF to be true.\n", which);
      DEBUG('c', " It is therfore conditionally waiting on full.\n\n", which);
      mutex->P();
      full->Wait();
      mutex->V();
    } // else
  } while(1);
} // method Producer

/*************************************************************************
 * Method: Consumer                                                      *
 * Purpose: Writes a character from the buffer to the screen.            *
 * Input: --PARAMETERS--                                                 *
 *        => which = ID value for the running thread.                    *
 * Output: --RETURNS--                                                   *
 *        => none                                                        *
 *************************************************************************/
void Consumer(int which)
{
  /* LOCAL VARIABLES */
  char charNextC;  // CHARACTER CONSUMED/WRITTEN

  do
  {
    /* MONITOR ENTRY */
    DEBUG('c', " Consumer '%dc' waiting on mutex.\n\n", which);
    mutex->P();

    /* CHECK EMPTINESS OF BUFFER */
    DEBUG('c', " Consumer '%dc' found Character Count vs. Buffer Size was '%d' vs. '%d'.\n\n", which, intCharCount,intBufferSize);
    if(intCharCount == 0)
    {
      DEBUG('c', " Consumer '%dc' conditionally waiting on empty.\n\n", which);
      empty->Wait();
    } // if

    /* GET CHARACTER FROM BUFFER */
    charNextC = charABuffer[intCIndex];
    printf("Consumer '%dc' consuming '%c'\n", which, charNextC);
    DEBUG('c', " The index read from was '%d'.\n\n", intCIndex);

    /* ADVANCE INDEX */
    intCIndex = (intCIndex + 1) % intBufferSize;

    /* DEC. NEW CHARACTER COUNT */
    intCharCount--;

    /* SIGNAL PRODUCER */
    DEBUG('c', " Consumer '%dc' conditionally signaling producer via full.\n\n", which);
    full->Signal();

    /* MONTIOR EXIT */
    if (next_count > 0)
    {
      DEBUG('c', " Consumer '%dc' signaling next.\n\n", which);
      next->V();
    } // if
    else
    {
      DEBUG('c', " Consumer '%dc' signaling mutex.\n\n", which);
      mutex->V();
    } // else
  } while(1);
} // method Consumer

/*************************************************************************
 * Method: ThreadTest                                                    *
 * Purpose: Sets up and controls the running the various producer and    *
 *          consumer threads.                                            *
 * Input: --PARAMETERS--                                                 *
 *        => none                                                        *
 * Output: --RETURNS--                                                   *
 *        => none                                                        *
 *************************************************************************/
void ThreadTest()
{
  /* LOCAL VARIABLES */
  char charAInputRead[30];    // INPUT CHARCTER READ
  int intProducers = 0;       // NUMBER OF PRODUCERS
  int intConsumers = 0;       // NUMBER OF CONSUMERS
  Thread *t;                  // THREAD POINTER FOR PRODUCER + CONSUMER
  int i;                      // LOOP COUNTER/ARRAY INDEX

  DEBUG('t', " Entering SimpleTest\n");

  /* GET AND CHECK NAME OF INPUT FILE */
  do
  {
    fflush(stdin);
    puts("Please enter the name of the file: ");
    gets(charAInputRead);
    puts("");
  } while(charAInputRead == NULL);
    
  DEBUG('c', " The given filename was '%s'.\n\n", charAInputRead);

  /* OPEN AND CHECK POINTER TO INPUT FILE */
  FileInput  = fopen(charAInputRead,"r");
    
  if(FileInput == NULL)
  {
    puts("FATAL ERROR: The file pointer was NULL.\n");
    exit(-1);
  } // if

  /* GET AND CHECK SIZE OF BUFFER */
  do
  {
    puts("Please enter the size of the buffer:");
    //scanf("%d",&intBufferSize);
    gets(charAInputRead);
    intBufferSize = atoi(charAInputRead);
    puts("");
  } while(intBufferSize < 1);

  DEBUG('c', " The size of the buffer given was '%d'.\n\n", intBufferSize);

  /* SET SIZE OF BUFFER */
  charABuffer = new char[intBufferSize];

  /*  GET AND CHECK NUMBER OF PRODUCERS */
  do
  {
    puts("Please enter the number of producers:");
    //scanf("%d",&intProducers);
    gets(charAInputRead);
    intProducers = atoi(charAInputRead);
    puts("");
  } while(intProducers < 1);

  DEBUG('c', " The number of producers given was '%d'.\n\n", intProducers);

  /* GET AND CHECK NUMBER OF CONSUMERS */
  do
  {
    puts("Please enter the number of consumers:");
    //scanf("%d",&intConsumers);
    gets(charAInputRead);
    intConsumers = atoi(charAInputRead);
    puts("");
  } while(intConsumers < 1);

  DEBUG('c', " The number of consumers given was '%d'.\n\n", intConsumers);
 
  /* INITIALIZE CONDITIONS */
  full = new Condition("full condition");
  empty = new Condition("empty condition");
  
  /* INITIALIZE SEMAPHORES */
  file = new Semaphore("file semaphore", 1);
  mutex = new Semaphore("mutex semaphore", 1);
  next = new Semaphore("next semaphore", 0);

  /* FORK THREADS */
  // => LAUNCH APPROPRIATE NUMBER OF PRODUCERS
  for(i = 0; i < intProducers; i++)
  {
    t = new Thread("producer thread");
    DEBUG('c', " Forking producer thread #'%d'.\n\n", i);
    t->Fork(Producer, i);
  } // for
    
  // => LAUNCH APPROPRIATE NUMBER OF CONSUMERS
  for(i = 0; i < intConsumers; i++)
  {
    t = new Thread("consumer thread");
    DEBUG('c', " Forking consumer thread #'%d'.\n\n", i);
    t->Fork(Consumer, i+intProducers);
  } // for
} // method ThreadTest
