/*********************************************************************
 * Authors: B. Alex Bridges and Miriam Becker                        *
 * File: smoker.cc/threadtest.cc                                     *
 * Class: CPSC-451, Summer 1999                                      *
 * Project: Project 4 - Cigarette Somker's Problem Solved via        *
 *                      Monitors (Implemented Using Semaphores)      *
 * Description: This program has an agent which places 2 of the 3    *
 *              ingredients used for cigarettes on the table, and 3  *
 *              different types of smokers, only one of which can    *
 *              use the ingredients to successfully smoke.           *
 * Contents: Methods for creating and controlling the agent and      *
 *           smokers as well as the actual code for the agent        *
 *           and smokers.                                            *
 *********************************************************************/
 
/* INCLUDES */
#include "copyright.h"
#include "system.h"
#include "synch.h"
#include "stdio.h"
#include "stdlib.h"

/* CONSTANTS */
const cTrue = 1;           // VALUE REPRESENTING THE TRUE CONDITION
const cFalse = 0;          // VALUE REPRESENTING THE FALSE CONDITION
const cPaper = 0;          // VALUE REPRESENTING THE INGREDIENT AS PAPER
const cTobacco = 1;        // VALUE REPRESENTING THE INGREDIENT AS TOBACCO
const cMatches = 2;        // VALUE REPRESENTING THE INGREDIENT AS MATCHES
//const cNothing = -1;       // VALUE REPRESENTING NO INGREDIENTS

/* GLOBAL VARIABLES */
int intMissIng = -1;       // STATUS AS TO INGREDIENT WHICH IS MISSING FROM TABLE
int bIngUsed = cTrue;      // STATUS AS TO WHETHER OR NOT INGREDIENTS HAVE BEEN USED
int intPaper = 1;          // NUMBER OF SMOKERS WITH PAPER
int intTobacco = 1;        // NUMBER OF SMOKERS WITH TOBACCO
int intMatches = 1;        // NUMBER OF SMOKERS WITH MATCH
int next_count = 0;        // NUMBER OF PROCEDURES THAT ARE IN MONITOR, BUT WAITING
// => CONDITIONS
Condition* agent;          // CONDITION FOR AGENT CONTROL
Condition* paper;          // CONDITION FOR CONTROLLING SMOKER WITH PAPER
Condition* tobacco;        // CONDITION FOR CONTROLLING SMOKER WITH TOBACCO
Condition* matches;        // CONDITION FOR CONTROLLING SMOKER WITH MATCHES
// => SEMAPHORES
Semaphore* mutex;          // SEMAPHORE FOR MUTEX CONTROL
Semaphore* next;           // SEMAPHORE FOR NEXT CONTROL


/* METHODS */
/*************************************************************************
 * Method: Agent                                                         *
 * Purpose: Places 2 of the 3 ingredients on the table, so that one of   *
 *          smoker types can smoke.                                      *
 *          pointer and places it in a rotating buffer.                  *
 * Input: --PARAMETERS--                                                 *
 *        => which = ID value for the running thread.                    *
 * Output: --RETURNS--                                                   *
 *        => none                                                        *
 *************************************************************************/
void Agent(int which)
{
  /* LOCAL VARIABLES */
  int intLoopCnt = -1; // NUMBER OF LOOPS
 
  do
  {
    /* DEBUG: ADJUST AND CHECK LOOP COUNT */
    //intLoopCnt++;
    //DEBUG('c', " Agent's loop count was '%d'.\n\n", intLoopCnt);
    //if(intLoopCnt == 10)
    //{
    //  DEBUG('c', " Agent breaking loop!!!\n\n", intLoopCnt);
    //  break;
    //} // if
    
    /* MONITOR ENTRY */
    DEBUG('c', " Agent waiting on mutex.\n\n");
    mutex->P();
      
    /* CHECK STATUS OF INGREDIENTS */
    // => INGREDIENTS NOT USED, SO WAIT
    if(bIngUsed == cFalse)
    {
      DEBUG('c', " Agent conditionally waiting on agent.\n\n");
      agent->Wait();
    } // if
    // => OTHERWISE CONTINUE
    else
    {
      /* DETERMINE NEW MISSING INGREDIENT AND SIGNAL APPROPRIATE SMOKER */
      intMissIng = (intMissIng + 1) % 3;
      switch(intMissIng)
      {
        case cPaper   : printf("Agent placing tobacco and matches on table.\n");
                        bIngUsed = cFalse;
                        DEBUG('c', " Agent conditionally signaling paper.\n\n");
                        paper->Signal();
                        break;
        case cTobacco : printf("Agent placing paper and matches on table.\n");
                        bIngUsed = cFalse;
                        DEBUG('c', " Agent conditionally signaling tobacco.\n\n");
                        tobacco->Signal();
                        break;
        case cMatches : printf("Agent placing paper and tobacco on table.\n");
                        bIngUsed = cFalse;
                        DEBUG('c', " Agent conditionally signaling matches.\n\n");
                        matches->Signal();
                        break;
      } // switch

      /* BLOCK UNTIL SMOKER SIGNALS ME = INGREDIANTS HAVE BEEN USED */
      //DEBUG('c', " Agent conditionally waiting on agent.\n\n");
      //agent->Wait();

      /* MONITOR EXIT */
      if (next_count > 0)
      {
        DEBUG('c', " Agent signaling next.\n\n");
        next->V();
      } // if
      else
      {
        DEBUG('c', " Agent signaling mutex.\n\n");
        mutex->V();
      } // else
    } // else
  } while(1);
} // method Agent

/*************************************************************************
 * Method: Smoker                                                        *
 * Purpose: General method which handles all 3 types of smokers.         *
 * Input: --PARAMETERS--                                                 *
 *        => which = ID value for the running thread.                    *
 * Output: --RETURNS--                                                   *
 *        => none                                                        *
 *************************************************************************/
void Smoker(int which)
{
  /* LOCAL VARIABLES */
  int intSmoker = -1;     // TYPE OF SMOKER AS INTEGER
  char charSmoker = '_';  // TYPE OF SMOKER AS CHARACTER

  /* DETERMINE TYPE OF SMOKER */
  // => SMOKER WITH PAPER
  if( (which >= 0) && (which < intPaper) )
  {
    intSmoker = cPaper;
    charSmoker = 'p';
    DEBUG('c', " The type for smoker '%d' is paper.\n\n", which);
  } // if
  // => SMOKER WITH TOBACCO
  else if( (which >= intPaper) && (which < intPaper+intTobacco) )
  {
    intSmoker = cTobacco;
    charSmoker = 't';
    DEBUG('c', " The type for smoker '%d' is tobacco.\n\n", which);
  } // else if
  // => SMOKER WITH MATCHES
  else if( (which >= intPaper+intTobacco) && (which < intPaper+intTobacco+intMatches) )
  {
    intSmoker = cMatches;
    charSmoker = 'm';
    DEBUG('c', " The type for smoker '%d' is matches.\n\n", which);
  } // else if
  // => UNKNOWN SMOKER
  else
    DEBUG('c', " The type for smoker '%d' could not be determined!\n\n");

  do
  {
    /* MONITOR ENTRY */
    DEBUG('c', " Smoker '%d%c' waiting on mutex.\n\n", which, charSmoker);
    mutex->P();
    
    /* CHECK STATUS OF MISSING INGREDIENT */
    // => MISSING INGREDIENT MATCHES ME AND NOT ALREADY BEEN USED, SO SMOKE AND SIGNAL AGENT
    if( (intMissIng == intSmoker) && (bIngUsed == cFalse) )
    {
      printf("Smoker '%d%c' rolling cigarette and smoking.\n\n", which, charSmoker);
      //intMissIng = cNothing;
      bIngUsed = cTrue;
      DEBUG('c', " Smoker '%d%c' signaling mutex.\n\n", which, charSmoker);
      mutex->V();
      DEBUG('c', " Smoker '%d%c' conditionally signaling agent.\n\n", which, charSmoker);
      agent->Signal();
    } // if
    // => OTHERWISE BLOCK UNTIL SIGNALED BY AGENT
    else
    {
      switch(intSmoker)
      {
        case cPaper   : 
          DEBUG('c', " Smoker '%d%c' conditionally waiting on paper.\n\n", which, charSmoker);
          paper->Wait();
          break;
        case cTobacco :
          DEBUG('c', " Smoker '%d%c' conditionally waiting on tobacco.\n\n", which, charSmoker);tobacco->Wait();
          break;
        case cMatches :
          DEBUG('c', " Smoker '%d%c' conditionally waiting on matches.\n\n", which, charSmoker);matches->Wait();
          break;
      } // switch
    } // else

    /* MONTIOR EXIT */
    if (next_count > 0)
    {
      DEBUG('c', " Smoker '%d%c' signaling next.\n\n", which, charSmoker);
      next->V();
    } // if
    else
    {
      DEBUG('c', " Smoker '%d%c' signaling mutex.\n\n", which, charSmoker);
      mutex->V();
    } // else
  } while(1);
} // method Smoker

/*************************************************************************
 * Method: ThreadTest                                                    *
 * Purpose: Sets up and controls running the agent and various types of  *
 *          smokers.                                                     *
 * Input: --PARAMETERS--                                                 *
 *        => none                                                        *
 * Output: --RETURNS--                                                   *
 *        => none                                                        *
 *************************************************************************/
void ThreadTest()
{
  /* LOCAL VARIABLES */
  int intSmokers = 0;  // TOTAL NUMBER OF SMOKERS
  Thread *t;           // THREAD POINTER FOR SMOKER
  int i;               // LOOP COUNTER/ARRAY INDEX

  DEBUG('t', " Entering SimpleTest\n");

  /*  GET AND CHECK NUMBER OF SMOKERS WITH PAPER */
  //do
  //{
  //  puts("Please enter the number of smokers with paper:");
  //  scanf("%d",&intPaper);
  //  puts("");
  //} while(intPaper < 1);

  DEBUG('c', " The number of smokers with paper given was '%d'.\n\n", intPaper);

  /*  GET AND CHECK NUMBER OF SMOKERS WITH TOBACCO */
  //do
  //{
  //  puts("Please enter the number of smokers with tobacco:");
  //  scanf("%d",&intTobacco);
  //  puts("");
  //} while(intTobacco < 1);

  DEBUG('c', " The number of smokers with tobacco given was '%d'.\n\n", intTobacco);

  /*  GET AND CHECK NUMBER OF SMOKERS WITH MATCHES */
  //do
  //{
  //  puts("Please enter the number of smokers with matches:");
  //  scanf("%d",&intMatches);
  //  puts("");
  //} while(intMatches < 1);

  DEBUG('c', " The number of smokers with matches given was '%d'.\n\n", intMatches);
 
  /* INITIALIZE CONDITIONS */
  agent = new Condition("agent condition");
  paper = new Condition("paper condition");
  tobacco = new Condition("tobacco condition");
  matches = new Condition("matches condition");
  
  /* INITIALIZE SEMAPHORES */
  mutex = new Semaphore("mutex semaphore", 1);
  next = new Semaphore("next semaphore", 0);

  /* FORK THREADS */
  // => LAUNCH APPROPRIATE NUMBER OF SMOKERS
  intSmokers = intPaper + intTobacco + intMatches;
  for(i = 0; i < intSmokers; i++)
  {
    t = new Thread("smoker thread");
    DEBUG('c', " Forking smoker thread #'%d'.\n\n", i);
    t->Fork(Smoker, i);
  } // for
  // => LAUNCH AGENT
  t = new Thread("agent thread");
  DEBUG('c', " Forking agent thread #'-1'.\n\n");
  t->Fork(Agent, -1);
} // method ThreadTest
